/*
 *  Uzebox Kernel
 *  Copyright (C) 2008-2009 Alec Bourque
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Uzebox is a reserved trade mark
*/

#include <avr/io.h>
#include <defines.h>

/*
.global InitSound
.global update_sound_buffer
.global update_sound_buffer_2
.global update_sound_buffer_fast
.global MixSound
.global SetMixerNote
.global SetMixerVolume
.global SetMixerWave
.global SetMixerNoiseParams
.global EnableSoundEngine
.global DisableSoundEngine
.global waves
.global mix_pos
.global mix_buf
.global mix_bank
.global tr4_barrel_lo
.global tr4_barrel_hi
.global tr4_params
.global sound_enabled
.global update_sound
.global counter
.global mixer

#define vol 			0
#define step_lo			1
#define step_hi			2
#define samplepos_frac	3
#define samplepos_lo	4
#define samplepos_hi	5
*/

#if SOUND_MIXER == MIXER_TYPE_INLINE
	#include "soundMixerInline.s"
#else
	#include "soundMixerVsync.s"
#endif	

.section .text

;**********************
; SET NOTE
;(C-call compatible)
; r24: Channel (0,1,2,3)
; r22: -MIDI note, 69=A5(440) for waves channels (0,1,2)
;      -Noise params for channel 3
;***********************
SetMixerNote:
	clr r25
	clr r23

#if MIXER_CHAN4_TYPE == 0
	cpi r24,3
	brlo set_note_waves
	ret
#endif

set_note_waves:
	;get step for note
	ldi ZL,lo8(steptable)
	ldi ZH,hi8(steptable)
	lsl r22
	rol r23
	add ZL,r22
	adc ZH,r23	

	lpm r26,Z+
	lpm r27,Z

	ldi ZL,lo8(mixerStruct)
	ldi ZH,hi8(mixerStruct)
	ldi r18,CHANNEL_STRUCT_SIZE
	mul r18,r24
	add ZL,r0
	adc ZH,r1
	
	std Z+step_lo,r26
	std Z+step_hi,r27
	
	clr r1
	

	ret

#if MIXER_CHAN4_TYPE == 0
;**********************
; SET NOISE CHANNEL PARAMS
;(C-call compatible)
; r24: noise divider
;*****************
SetMixerNoiseParams:
	;preserve wave type (7/15 bit)
	lds r25,tr4_params
	andi r25,1
	lsl r24
	or r24,r25

	sts tr4_params,r24	
	ret

#endif

;**********************
; SET sound patch for channel 
; (C-call compatible)
; r24 channel (0,1,2)
; r23:r22 Waves channels: patch (0x00-0xfd) 
;         Noise channel: 0xfe=7 bit lfsr, 0xff=15 bit lfsr
;                 
;***********************
SetMixerWave:
	clr r25
	clr r23

	ldi ZL,lo8(mixerStruct)
	ldi ZH,hi8(mixerStruct)
	ldi r18,CHANNEL_STRUCT_SIZE
	mul r18,r24	
	add ZL,r0
	adc ZH,r1

#if MIXER_CHAN4_TYPE == 0
	cpi r22,0xfe	;7bit lfsr
	brne smw1
	lds r22,tr4_params
	andi r22,0xfe;
	sts tr4_params,r22
	rjmp esmw	
smw1:
	cpi r22,0xff	;15bit lfsr
	brne smw2
	lds r22,tr4_params
	ori r22,0xfe;
	sts tr4_params,r22	
	rjmp esmw
smw2:
#endif

	ldi r23,hi8(waves)
	add r23,r22
	std Z+samplepos_hi,r23 ;store path No

esmw:
	clr r1	
	ret




;**********************
; SET Volume
; (C-call compatible)
; r24 channel (0,1,2,3)
; r22 volume (0x00-0xff)
;***********************
SetMixerVolume:
	clr r25
	clr r23

	ldi ZL,lo8(mixerStruct)
	ldi ZH,hi8(mixerStruct)
	ldi r18,CHANNEL_STRUCT_SIZE
	mul r18,r24	
	add ZL,r0
	adc ZH,r1
	std Z+vol,r22 ;store vol

	clr r1	
	ret


;*****************************
; Enabled the sound engine (mixed and player).
; C-callable
;*****************************
.section .text.EnableSoundEngine
EnableSoundEngine:	
	ldi r24,1
	sts sound_enabled,r24
	ret

;*****************************
; Disable the sound engine. When disabled, 
; no CPU cycles are consumed.
; C-callable
;*****************************
.section .text.DisableSoundEngine
DisableSoundEngine:	
	ldi r24,0
	sts sound_enabled,r24
	ret


steptable:
#include "data/steptable.inc"

.align 8
waves:
#if INCLUDE_DEFAULT_WAVES == 1
	#include MIXER_WAVES
#endif








